-- All code written and maintained by Yewchi 
-- zyewchi@gmail.com

--- Initialization --
--
------- Flags --
--
--
------ /Flags --

------- Table key references --
--
--
------ /Table key references --

------- Static vals --
--
local intToHexTable = {[10] = 'A', [11] = 'B', [12] = 'C', [13] = 'D', [14] = 'E', [15] = 'F'}
local hexToIntTable = {['A'] = 10, ['B'] = 11, ['C'] = 12, ['D'] = 13, ['E'] = 14, ['F'] = 15}

local SAFETY_UTF_BUMP = 0x1B
local SECRET_NUMBER = 0x31

local LINK_START_SPELL = "flexi$"
local LINK_START_ITEM = "flexi#"

local SPELL_LINK_COLOUR = "|cff71d5ff"
--
------- /Static vals --

------- Return values --
--
--
------ /Return values --

------- AddOn Variables --
--
local runThisFlexiDecodeInCompat = false
--
------ /AddOn Variables --
--
-- /Initialization --



--- Funcs --
--
local function round(val, nearest)
	return floor((val + (nearest / 2)) / nearest)
end

local function GetHex(num)
	assert(num < 16)
	if num >= 10 then
		return intToHexTable[num]
	end
	return num
end

local function GetInt(hex)
	local num = tonumber(hex)
	if type(num) == "number" then
		return num
	end
	return hexToIntTable[hex:upper()]
end

local function IntToHex256(num)
	local num = max(0, min(255, num))
	local high = floor(num / 16)
	local low = (num % 16)
	
	return GetHex(high)..GetHex(low)
end

local function HexToInt256(str)
	local multiplier = 1
	local total = 0
	
	for i=#str,1,-1 do
		local c = str:sub(i, i)
		total = total + (multiplier * GetInt(c))
		multiplier = multiplier * 16
	end
	
	return total
end

local function CompressColour(noAlphaIntColour)
	local total = 0
	local modValMultiplier = 3
	local divValMultiplier = 2
	local thisMagnitudesBitNumber = 8
	
	while divValMultiplier >= 0 do
		local thisMagnitudesValue = noAlphaIntColour % (256^modValMultiplier)

		thisMagnitudesValue = floor(thisMagnitudesValue / (256^(divValMultiplier)))

		if divValMultiplier == 1 then
			thisMagnitudesValue = min(3, round(thisMagnitudesValue, 64)) -- put the 8 bit number into TWO bits
			thisMagnitudesBitNumber = thisMagnitudesBitNumber - 2 -- Drop to 4th bit
		else
			thisMagnitudesValue = min(7, round(thisMagnitudesValue, 32)) -- put the 8 bit number into THREE bits
			thisMagnitudesBitNumber = thisMagnitudesBitNumber - 3 -- Drop to {6th, 1st} bit
		end

		total = total + (thisMagnitudesValue * (2^(thisMagnitudesBitNumber)))
		
		modValMultiplier = modValMultiplier - 1
		divValMultiplier = divValMultiplier - 1
	end
	
	return string.char(total)
end

local function DecompressColour(char)
	local asciiCode = string.byte(char)
	local hexColourString = ""
	
	local modValMultiplier = 8
	local divValMultiplier = 5
	
	while divValMultiplier >= 0 do
		local thisMagnitudesValue = asciiCode % (2^modValMultiplier)
		thisMagnitudesValue = floor(thisMagnitudesValue / (2^divValMultiplier))
		
		if divValMultiplier == 3 then
			thisMagnitudesValue = thisMagnitudesValue * (2^6) + 32 -- Expand to 8 bits
			modValMultiplier = modValMultiplier - 2
			divValMultiplier = divValMultiplier - 3
		else
			thisMagnitudesValue = thisMagnitudesValue * (2^5) + 16 -- Expand to 8 bits
			modValMultiplier = modValMultiplier - 3
			divValMultiplier = divValMultiplier - 2
		end

		hexColourString = string.format("%s%s", hexColourString, IntToHex256(thisMagnitudesValue))
	end
	
	return hexColourString
end

local function COMPATIBILITY10_SixBitCharsToNumber(str)
	local total = 0
	local remade=""
	local magnitude = #str - 1
	local sixthPower = 2^6
	local fifthPower = 2^5 -- Will return the 7th bit to it's original position
	
	for c in str:gmatch("(.)") do
		local charValue = string.byte(c)
		total = total + ((floor(charValue / sixthPower) * fifthPower) + (charValue % fifthPower)) * 64^magnitude
		magnitude = magnitude - 1
	end
	return total
end

---- Base_64 numbers, for compression into ascii codes --
--
local function NumberToSixBitChars(num)
	assert(type(num) == "number")
	
	local fullNumber = num
	local chars = ""
	local fifthPower = 2^5
	
	-- Re: "this6Bits = ... + (this6Bits < 32 and fifthPower or 0)"..
	-- The addition of fifthPower is due to numbers below 32 being invalid UTF-8 character's codes, and when the 
	-- top bit of a 6-bit number is bumped up to the 7th bit, we gain [0 or 32]-"value space", keeping the count 
	-- upwards 31->32..->63 consistent through the bit jump
	while num > 0 do
		local this6Bits = num % 64
		this6Bits = (this6Bits % fifthPower) + (floor(this6Bits / fifthPower) * fifthPower * 2) + (this6Bits < 32 and fifthPower or 0) + SAFETY_UTF_BUMP
		
		chars = string.format("%s%s", string.char(this6Bits), chars)
		
		num = floor(num / 64)
	end
	
	return chars
end

local function SixBitCharsToNumber(str)
	assert(str)
	
	local total = 0
	local remade=""
	local magnitude = #str - 1
	local sixthPower = 2^6
	local fifthPower = 2^5 -- Will return the 6th, 7th bits to their original position
	
	local charTbl = {}
	
	if runThisFlexiDecodeInCompat ~= false then
		if runThisFlexiDecodeInCompat == "1.0" then
			return COMPATIBILITY10_SixBitCharsToNumber(str)
		end
	end
	
	for c in str:gmatch("(.)") do -- Builds table of UTF-8 code points from each character, and checks if the code must be from an older version (not always deducable, but could save issues with players running an older version.
		local thisCharValue = string.byte(c)
		if thisCharValue < 0x3B or thisCharValue > 0x7A then -- Ensures if a Flexi-v1.0 code is found it will be interpreted as the old character set.
			runThisFlexiDecodeInCompat = "1.0"
			return COMPATIBILITY10_SixBitCharsToNumber(str)
		end
		
		table.insert(charTbl, thisCharValue)
	end
	
	for _,charValue in ipairs(charTbl) do
		charValue = charValue - SAFETY_UTF_BUMP
		charValue = charValue - (charValue < 64 and fifthPower or 0)
		
		total = total + ((floor(charValue / sixthPower) * fifthPower) + (charValue % fifthPower)) * 64^magnitude
		
		magnitude = magnitude - 1
	end
	
	return total
end

--
----  Base_64 --
function FLEXI.DecodeFlexToItemLink(flexCode)
	local itemStringValsFinder = flexCode:gmatch("([^:]*):")
	itemStringVals = {}
	
	local i = 0
	for val in itemStringValsFinder do
		if i > 0 then 
			itemStringVals[i] = SixBitCharsToNumber(val)
		end
		i = i + 1
	end
	local itemInfo = {GetItemInfo(itemStringVals[FLEXI.IS_ID])}

	local itemName = itemInfo[1]
	
	local decodedItemLinkFinder = string.gmatch(itemInfo[2]..":", "([^:]*):")
	
	local i = 1
	local decodedItemLinkVals = {}
	for val in decodedItemLinkFinder do
		if i > 0 then 
			local replacingVal = itemStringVals[i-1]
			decodedItemLinkVals[i] = (replacingVal ~= nil and replacingVal ~= 0) and itemStringVals[i-1] or val
		else
			decodedItemLinkVals[i] = val
		end
		i = i + 1
	end
	
	local decodedItemLink = decodedItemLinkVals[1]
	for i=2,#decodedItemLinkVals,1 do
		decodedItemLink = string.format("%s:%s", decodedItemLink, decodedItemLinkVals[i])
	end
	
	runThisFlexiDecodeInCompat = false
	
	return decodedItemLink
end

function FLEXI.EncodeItemLinkToFlex(str)
	local generatedFlexCode = LINK_START_ITEM
	
	local itemLinkData = str:gmatch("([^:]*):")
	local itemId, enchantId, gemId1, gemId2, gemId3, gemId4, suffixId, uniqueId, linkLevel, 
			specId, upgradeTypeId, instanceDifficultyId = itemLinkData
		
	local i = 1
	for sub in itemLinkData do
		if i > 12 then 
			generatedFlexCode = string.format("%s:", generatedFlexCode)
		
			return generatedFlexCode -- Valid return value
		end
		
		local checkNum = tonumber(sub)
		if type(checkNum) == "number" then 
			status, sub = pcall(NumberToSixBitChars, checkNum)
			if not status then
				--print("error on "..i)
			end
		end
		generatedFlexCode = string.format("%s:%s", generatedFlexCode, sub)
		--print("itemLink["..i.."] = "..sub)
		i = i + 1
	end
	
	return nil
end

-- Operates differently from items, because the flexCode is only sending the spellId (which was deduced from the message sender, and everything else can be deduced, i.e.: all info about a spell can be known from a spellId. I'm not sure how talent/item modifiers work with sending/receiving spell links, yet.)
function FLEXI.DecodeFlexToSpellLink(flexCode)
	local spellStringValsFinder = flexCode:gmatch("([^:]*):")
	local spellStringVals = {}
	
	local i = 0
	for val in spellStringValsFinder do
		if i > 0 then 
			spellStringVals[i] = val
		end
		i = i + 1
	end
	
	local spellId = SixBitCharsToNumber(spellStringVals[1])
	local spellSubtext = GetSpellSubtext(spellId)
	local spellName = GetSpellInfo(spellId)
	
	-- If the flexiCode passes the (simple) maths test...
	if (spellId+SECRET_NUMBER) % 64 == SixBitCharsToNumber(spellStringVals[2]) then
		if spellSubtext ~= nil and spellSubtext ~= "" then
			local spellRank = string.format(" (%s)", spellSubtext)
			spellName = string.format("%s%s", spellName, spellRank)
		end
		
		if spellName ~= nil then
			local decodedSpellLink = string.format("%s|Hspell:%d|h[%s]|h|r", SPELL_LINK_COLOUR, spellId, spellName)
		
			return decodedSpellLink
		end
	end
	
	return nil
end

function FLEXI.EncodeSpellIdToFlex(spellId)
	if spellId ~= nil then
		return string.format("%s:%s:%s:", LINK_START_SPELL, NumberToSixBitChars(spellId), NumberToSixBitChars((spellId+SECRET_NUMBER) % 64))
	end
	
	return nil
end
--
-- /Funcs --